unit DataUnit;

{$mode ObjFPC}{$H+}
{$MODESWITCH ADVANCEDRECORDS}

// -----------------------------------------------------------
// Low level file access unit "poor man's file based database"
// (C)2015-2021 arminlinder@arminlinder.de

// Follows the CRUD pattern.
// Note: since the CRUD "create" clashes with TObject.Create, we'll use the "append" keyword

// V2.0 initial release, V1.0 plus a generic interface
// V2.1 support for pseudo-deletion and background refresh
// V2.2 Error handler event
// -----------------------------------------------------------

// Note on deleted records
// -----------------------
// Due to the simplistic multitasking approach used, and data security concerns,
// records are never physically deleted, but only marked for re-use.
// In addition, they are random-filled (if encryption is enabled) or zeroed or padded with the record number (if DEBUG)

// Todo
//   assert FDataCache.Count = inherited CountRec in the facade
//   recycle deleted records

interface

uses
  Classes, SysUtils, fgl, BlowFish, FileUtil, Forms, HexDumpUnit;

type

  { TBaseFileAccessor }

  // ----------------------------------------------------------
  // The core Accessor Object - low level file access interface
  // ----------------------------------------------------------

  TBaseFileAccessor = class(TObject)
  private
    FEncryptKey: string;
    FFileLastLoaded: TDateTime;
    FFilePath: string;
    FisEncrypted: boolean;
    FRecSize: integer;
    FFileStream: TFileStream;
    FDecryptStream: TBlowFishDecryptStream;
    FEncryptStream: TBlowFishEncryptStream;
    procedure SetEncryptKey(AValue: string);
    procedure SetFileLastLoaded(AValue: TDateTime);
    procedure SetFilePath(AValue: string);
    procedure SetisEncrypted(AValue: boolean);
  protected
    // File stream handling
    procedure OpenFileStreams;
    procedure CloseFileStreams;
    // CRUD (Low level to file)
    function AppendRec(const PBuffer: Pointer): integer;   // Append a new recorf
    function ReadRec(const PBuffer: Pointer; Idx: integer): boolean;       // Read Record [Idx]
    function UpdateRec(const PBuffer: Pointer; Idx: integer): boolean;     // Update record [Idx] Note: must be created first!
    function DeleteRec(Idx: integer): boolean;
  public
    // Encryption
    property isEncrypted: boolean read FisEncrypted write SetisEncrypted;
    property EncryptKey: string read FEncryptKey write SetEncryptKey;
    // Refresh functionality
    function isDirty: boolean;
    property FileLastLoaded: TDateTime read FFileLastLoaded write SetFileLastLoaded;
    // Helpers
    function HexDumpRec(Idx: integer): string;                // return a hex-dump of the record
    property FilePath: string read FFilePath write SetFilePath; // path of the database table file
    property RecSize: integer read FRecSize; // record size, set this in the constructor
    function CountRec: integer; // total number of records in the file
    function inRange(AValue, Min, Max: integer): boolean; // check if index is in valid range
    // Constructor
    constructor Create(AFileName: string; ARecSize: integer); // create a new table
  end;

  // ----------------------------------------------------------
  // The Generic Accessor Interface
  // ----------------------------------------------------------

  { Refresh Notify Event }
  TFileAccessorNotifyEvent = procedure(Sender: TObject; idx: integer) of object;

  { Error Notify Event }
  TFileAccessorErrorEvent = procedure(AClassName, ARoutineName, AFilePath: string; InIdx: integer; AMessage: string) of object;

  { TFileAccessor }
  generic TFileAccessor<T> = class(TBaseFileAccessor)

  type TDataCache = specialize TFPGList<T>;

  private
    FDataCache: TDataCache; // generic list used to cache file accesses
    FisRefreshing: boolean;
    procedure SetisRefreshing(AValue: boolean);
  public
    // events
    onAppended: TFileAccessorNotifyEvent;  // Element has been successfully added to file and cache
    onRead: TFileAccessorNotifyEvent;  // Element has been successfully read
    onUpdated: TFileAccessorNotifyEvent;   // Element has been successfully updated on file and in cache
    onDeleted: TFileAccessorNotifyEvent;   // Element has been marked delete
    onRefreshStart: TNotifyEvent; // an update cycle has started
    onRefreshEnd: TNotifyEvent;   // an update cycle has ended
    onRefreshed: TFileAccessorNotifyEvent; // a record has been replaced by a newer version
    onSkipped: TFileAccessorNotifyEvent; // a record has been checked and found not modified
    // Error handler event
    onError: TFileAccessorErrorEvent; // an error has occurred
  public  // properties
    property isRefreshing: boolean read FisRefreshing write SetisRefreshing;
  public
    function Append(const ARec: T): integer; overload;  // Append ARec
    function Append(): integer; overload;               // Append an empty record
    function Read(var ARec: T; idx: integer): boolean;  // Read record [idx] from cache
    function Delete(idx: integer): boolean;             // Mark record [idx] deleted
    function Update(ARec: T; idx: integer): boolean;    // overwrite ARec
    function Refresh: boolean;
    function InitRec: T;                                // Return a random initialized record
    constructor Create(AFilePath: string);
    destructor Destroy; override;
  end;

// ------------------------------------------------------------------------------------------
// ---------------------------------- Implementation ----------------------------------------
// ------------------------------------------------------------------------------------------

implementation

uses Unit1;

{ TBaseFileAccessor }

// ----------------------------
// CRUD routines
// ----------------------------

function TBaseFileAccessor.AppendRec(const PBuffer: Pointer): integer;
  // append record, return index

begin
  Result := -1;
  OpenFileStreams;
  try
    FFileStream.Seek(0, soEnd);
    if isEncrypted then
      FEncryptStream.WriteBuffer(PBuffer^, FRecSize)
    else
      FFileStream.WriteBuffer(PBuffer^, FRecSize);
  finally
    CloseFileStreams;
    Result := Self.CountRec - 1;
  end;
end;

function TBaseFileAccessor.ReadRec(const PBuffer: Pointer; Idx: integer): boolean;
  // Read existing record

begin
  Result := False;
  if not inRange(Idx, 0, Self.CountRec - 1) then exit;
  OpenFileStreams;
  try
    FFileStream.Seek(Idx * FRecSize, soBeginning);
    if isEncrypted then
      FDecryptStream.ReadBuffer(PBuffer^, FRecSize)
    else
      FFileStream.ReadBuffer(PBuffer^, FRecSize);
    Result := True;
  finally
    CloseFileStreams;
  end;
end;

function TBaseFileAccessor.UpdateRec(const PBuffer: Pointer; Idx: integer): boolean;

  // overwrite an existing record

begin
  Result := False;
  if not inRange(Idx, 0, Self.CountRec - 1) then exit;
  OpenFileStreams;
  try
    FFileStream.Seek(Idx * FRecSize, soBeginning);
    if isEncrypted then
      FEncryptStream.WriteBuffer(PBuffer^, FRecSize)
    else
      FFileStream.WriteBuffer(PBuffer^, FRecSize);
    Result := True;
  finally
    CloseFileStreams;
  end;
end;

function TBaseFileAccessor.DeleteRec(Idx: integer): boolean;

begin
  // delete an existing record
  assert(False, Format(Self.ClassName + {$I %CURRENTROUTINE%} + ': %s',
    ['Low-Level record delete is not allowed. Write update a record with the deleted flag set instead.']));
  Result := False;
end;

function TBaseFileAccessor.isDirty: boolean;

var
  TimeStamp: TDateTime;

begin
  Result := False;
  assert(FilePath <> '', Self.ClassName + '.' + {$I %CURRENTROUTINE%} + ': file name not set');
  if FileExists(Self.FilePath) then
  begin
    Timestamp := FileDateToDateTime(FileAge(Self.FilePath));
    Result := TimeStamp <> Self.FileLastLoaded;
  end;
end;

// ---------------------
// Misc helper functions
// ---------------------

function TBaseFileAccessor.CountRec: integer;
  // return the calculated number of records in the file
  // return 0 if the file does not exist

var
  s: int64;

begin
  s := FileUtil.FileSize(FilePath);
  if s = -1 then exit(0);  // not existing
  // check if file size is a multiple of recsize
  if (s mod FRecSize) <> 0 then raise Exception.Create(Format('%s.%s: invalid record size: %d', [Self.ClassName, {$I %CURRENTROUTINE%}, FRecSize]));
  // calculate number of records in the file
  Result := s div FRecSize;
end;

function TBaseFileAccessor.inRange(AValue, Min, Max: integer): boolean;
  // check if index is in given range
begin
  Result := False;
  if AValue > Max then exit;
  if AValue < Min then exit;
  Result := True;
end;

// ----------------------------
// Property setters and getters
// ----------------------------

procedure TBaseFileAccessor.SetEncryptKey(AValue: string);
begin
  if FEncryptKey = AValue then Exit;
  FEncryptKey := AValue;
end;

procedure TBaseFileAccessor.SetFileLastLoaded(AValue: TDateTime);
begin
  if FFileLastLoaded = AValue then Exit;
  FFileLastLoaded := AValue;
end;

procedure TBaseFileAccessor.SetFilePath(AValue: string);
begin
  if FFilePath = AValue then Exit;
  FFilePath := AValue;
end;

procedure TBaseFileAccessor.SetisEncrypted(AValue: boolean);
begin
  if FisEncrypted = AValue then Exit;
  FisEncrypted := AValue;
end;

// ----------------------
// File Stream management
// ----------------------

procedure TBaseFileAccessor.OpenFileStreams;

begin
  assert(not assigned(FFileStream), Format(Self.ClassName + {$I %CURRENTROUTINE%} + ': %s (%s)', ['FFileStream already open', FFilePath]));
  if FileExists(FFilePath) then
    FFileStream := TFileStream.Create(FFilePath, fmOpenReadWrite or fmShareDenyWrite)
  else
    FFileStream := TFileStream.Create(FFilePath, fmCreate or fmShareDenyWrite);
  if isEncrypted then
  begin
    assert(not assigned(FEncryptStream), Format(Self.ClassName + {$I %CURRENTROUTINE%} + ': %s', ['FEncryptStream already open']));
    assert(not assigned(FDecryptStream), Format(Self.ClassName + {$I %CURRENTROUTINE%} + ': %s', ['FDecryptStream already open']));
    FEncryptStream := TBlowFishEncryptStream.Create(FEncryptKey, FFileStream);
    FDecryptStream := TBlowFishDecryptStream.Create(FEncryptKey, FFileStream);
  end;
end;

procedure TBaseFileAccessor.CloseFileStreams;

begin
  if assigned(FDecryptStream) then FreeAndNil(FDecryptStream);
  if assigned(FEncryptStream) then FreeAndNil(FEncryptStream);
  if assigned(FFileStream) then FreeAndNil(FFileStream);
end;

// ---------------
// Diagnostics aid
// ---------------

function TBaseFileAccessor.HexDumpRec(Idx: integer): string;
  // simple hex-dump of an existing record as a string
  // feel free to use HexDumpToMemo of the HexDumpUnit for formatted output to a memo instead
var
  Buffer: Pointer;

begin
  Result := Format('Record[%d]: (%d)Bytes%s', [Idx, FRecSize, LineEnding]);
  if Idx = -1 then exit(Result + '#UNDEFINED');
  if not inRange(Idx, 0, Self.CountRec - 1) then exit(Result + '#RANGEERROR');
  GetMem(Buffer, FRecSize);
  try
    Self.ReadRec(Buffer, idx);
    Result := Result + HexDump(Buffer^, FRecSize);
  finally
    FreeMem(Buffer);
  end;
end;

// -------------------------
// Class constructor
// -------------------------

constructor TBaseFileAccessor.Create(AFileName: string; ARecSize: integer);

begin
  assert(ARecSize > 0, Format('%s.%s: invalid record size: %d', [Self.ClassName, {$I %CURRENTROUTINE%}, ARecSize]));
  assert(AFileName <> '', Format('%s.%s: empty file name', [Self.ClassName, {$I %CURRENTROUTINE%}]));
  FFilePath := AFileName;
  FisEncrypted := False;
  FFileLastLoaded := 0;
  FEncryptKey := '';
  FRecSize := ARecSize;
end;

// --------------------------------------------------------------------------------
// --------------------------------- Interface part -------------------------------
// --------------------------------------------------------------------------------

{ TFileAccessor }

procedure TFileAccessor.SetisRefreshing(AValue: boolean);
begin
  if FisRefreshing = AValue then Exit;
  FisRefreshing := AValue;
end;

function TFileAccessor.Append(const ARec: T): integer; overload;

var
  Buffer: T;
  idx: integer;

begin
  idx := -1;
  Buffer := ARec;
  Buffer.isDeleted := False;
  Buffer.modifiedAt := now();
  try
    idx := inherited AppendRec(@Buffer);
    if (idx > -1) then
    begin
      FDataCache.Add(Buffer);
      if assigned(onAppended) then
        onAppended(Self, idx);
    end;
  except
    on e: Exception do
    begin
      if assigned(onError) then
        onError(Self.ClassName, {$I %CURRENTROUTINE%}, Self.FilePath, idx, e.Message)
    end;
  end;
  Result := idx;
end;

function TFileAccessor.Append(): integer; overload;

begin
  Result := Append(Self.InitRec());
end;

function TFileAccessor.Read(var ARec: T; idx: integer): boolean;

begin
  Result := False;
  assert(inRange(idx, 0, FDataCache.Count - 1), Format(Self.ClassName + {$I %CURRENTROUTINE%} + ': %s %d', ['index out of range: ', idx]));
  try
    ARec := FDataCache[idx];
    if assigned(onRead) then
      onRead(Self, idx);
    Result := True;
  except
    on e: Exception do
    begin
      if assigned(onError) then
        onError(Self.ClassName, {$I %CURRENTROUTINE%}, Self.FilePath, idx, e.Message);
    end;
  end;
end;

function TFileAccessor.Update(ARec: T; idx: integer): boolean;

var
  Buffer: T;

begin
  Result := False;
  assert(inRange(idx, 0, FDataCache.Count - 1), Format(Self.ClassName + {$I %CURRENTROUTINE%} + ': %s %d', ['index out of range: ', idx]));
  Buffer := ARec;
  Buffer.ModifiedAt := Now();
  try
    if UpdateRec(@Buffer, idx) then
    begin
      FDataCache[idx] := Buffer;
      if assigned(onUpdated) then
        onUpdated(Self, idx);
      Result := True;
    end;
  except
    on e: Exception do
    begin
      if assigned(onError) then
        onError(Self.ClassName, {$I %CURRENTROUTINE%}, Self.FilePath, idx, e.Message);
    end;
  end;
end;

function TFileAccessor.Delete(idx: integer): boolean;

var
  Buffer: T;

begin
  Result := False;
  assert(inRange(idx, 0, FDataCache.Count - 1), Format(Self.ClassName + {$I %CURRENTROUTINE%} + ': %s %d', ['index out of range: ', idx]));
  Buffer := Self.InitRec();
  Buffer.isDeleted := True;
  try
    if Self.Update(Buffer, idx) then
    begin
      FDataCache[idx] := Buffer;
      if assigned(onDeleted) then
        onDeleted(Self, idx);
      Result := True;
    end;
  except
    on e: Exception do
    begin
      if assigned(onError) then
        onError(Self.ClassName, {$I %CURRENTROUTINE%}, Self.FilePath, idx, e.Message);
    end;
  end;
end;

// -------------------------------------------------------------
// --------------------- File refresh --------------------------
// -------------------------------------------------------------

function TFileAccessor.Refresh: boolean;

var
  PBuffer: ^T = nil;
  dt: TDateTime;
  idx: integer;

begin
  Result := False;
  if not isRefreshing then
  begin
    isRefreshing := True;
    idx := -1;
    if FileExists(Self.FilePath) then  // Sync file contents by timestamp
    begin
      if IsDirty then
      begin
        if assigned(onRefreshStart) then onRefreshStart(self);
        try
        OpenFileStreams;
        try
          dt := FileDateToDateTime(FileAge(FilePath));
          GetMem(PBuffer, FRecSize);
          while FFileStream.Position < FFileSTream.Size do
          begin
            Inc(idx);
            if isEncrypted then
              FDecryptStream.ReadBuffer(PBuffer^, FRecSize)
            else
              FFileStream.ReadBuffer(PBuffer^, FRecSize);
            if idx < (FDataCache.Count) then  // modify existing
            begin
              if PBuffer^.ModifiedAt > FDataCache[idx].modifiedAt then
              begin
                FDataCache[idx] := PBuffer^;
                if PBuffer^.isDeleted then
                  if assigned(onDeleted) then
                    onDeleted(Self, idx)
                  else
                  if assigned(onRefreshed) then
                    onRefreshed(Self, idx);
                Result := True;
              end
              else
              begin
                if assigned(onSkipped) then
                  onSkipped(Self, idx);
              end;
            end
            else  // append new
            begin
              FDataCache.Add(PBuffer^);
              Result := True;
              if assigned(onAppended) then
                onAppended(Self, idx);
            end;
          end;      // while not eof
          FileLastLoaded := dt;
          if assigned(onRefreshEnd) then onRefreshEnd(self);
          Result := True;
        finally
          if assigned(PBuffer) then FreeMem(PBuffer);
          CloseFileStreams;
        end;
        except
          on e: Exception do
          begin
            if assigned(onError) then
              onError(Self.ClassName, {$I %CURRENTROUTINE%}, Self.FilePath, -1, e.Message)
          end;
         end; // except
      end;  // is Direty
    end;  // file exists
    // truncate FDataCache if file is smaller or missing
    for idx := FDataCache.Count - 1 downto Self.CountRec do
    begin
      FDataCache.Delete(idx);
      if assigned(onDeleted) then
        onDeleted(Self,idx);
      Result := True;
    end;
    isRefreshing := False;
  end; // is Refreshing
end;

function TFileAccessor.InitRec: T;

var
  ArrBuffer: array[0..$FFFF] of byte absolute Result;
  i: integer;

begin
  // Initialize an empty record
  if isEncrypted then       // init the record with random bytes
  begin
    for i := 0 to FRecSize - 1 do
      ArrBuffer[i] := Random(256);
  end
  else
    {$IFOPT D+}
       FillChar(ArrBuffer, FRecSize, lo(Self.CountRec));   // just a little aid to find the record data in a hex stream
    {$ELSE}
      FillChar(ArrBuffer, FRecSize, $00);   // just a little aid to find the record data in a hex stream
    {$ENDIF}
  Result.isDeleted := False;
  Result.ModifiedAt := Now();
end;


// -------------------------------------------
// TFileAccessor Constructor and Destructor
// -------------------------------------------

constructor TFileAccessor.Create(AFilePath: string);

begin
  inherited Create(AFilePath, SizeOf(T));
  FDataCache := TDataCache.Create;
  fIsRefreshing := False;
end;

destructor TFileAccessor.Destroy;

begin
  FreeAndNil(FDataCache);
  inherited Destroy;
end;

end.
