unit Unit1;

{$mode objfpc}{$H+}
{$MODESWITCH ADVANCEDRECORDS}

// ----------------------------------
// Test program for the DataFile unit
// (c)2021 arminlinder@arminlinder.de
// ----------------------------------

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls,
  ComCtrls, ExtCtrls, Spin, LazFileUtils, DataUnit;

type

  { TForm1 }

  TForm1 = class(TForm)
    ButtonMemoClear: TButton;
    ButtonAppend: TButton;
    ButtonDelete: TButton;
    CheckBoxEncrypted: TCheckBox;
    CheckBoxAutoRefresh: TCheckBox;
    EditLastRefresh: TEdit;
    LabeledEditPassword: TLabeledEdit;
    Memo1: TMemo;
    MemoEvents: TMemo;
    RadioGroup1: TRadioGroup;
    SpinEdit1: TSpinEdit;
    SpinEditRefresh: TSpinEdit;
    StatusBar1: TStatusBar;
    TimerRefresh: TTimer;
    procedure ButtonMemoClearClick(Sender: TObject);
    procedure ButtonAppendClick(Sender: TObject);
    procedure ButtonDeleteClick(Sender: TObject);
    procedure CheckBoxEncryptedChange(Sender: TObject);
    procedure CheckBoxAutoRefreshChange(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure LabeledEditPasswordChange(Sender: TObject);
    procedure SpinEdit1Change(Sender: TObject);
    procedure TimerRefreshTimer(Sender: TObject);
  private
  public
    // event handlers
    procedure onAppended(Sender: TObject; idx: integer);
    procedure OnRefreshStart(Sender: TObject);
    procedure OnRefreshEnd(Sender: TObject);
    procedure OnRefreshed(Sender: TObject; idx: integer);
    procedure OnSkipped(Sender: TObject; idx: integer);
    procedure onRead(Sender: TObject; idx: integer);
    procedure onUpdated(Sender: TObject; idx: integer);
    procedure onDeleted(Sender: TObject; idx: integer);
    procedure onError(AClassName, ARoutineName, AFilePath: string; AnIdx: integer; AMessage: string);
    // helpers
    procedure DumpRecord(idx: integer);
    function CurrentAccessor: TBaseFileAccessor;
    procedure RefreshAll;
    // Event protocol
    procedure DisplayEvent(EventMsg: string);
  end;

// Data types for testing
type
  TRecordOne = record
    isDeleted: boolean;
    modifiedAt: TDateTime;
    Content: shortstring;
    class operator = (const A, B: TRecordOne): boolean;
  end;

  TRecordTwo = record
    isDeleted: boolean;
    modifiedAt: TDateTime;
    I: integer;
    class operator = (const A, B: TRecordTwo): boolean;
  end;

// data type specific accessors

type
  TRecordFileAccessorOne = class(specialize TFileAccessor<TRecordOne>);
  TRecordFileAccessorTwo = class(specialize TFileAccessor<TRecordTwo>);

var
  Form1: TForm1;
  RecordFileAccessorOne: TRecordFileAccessorOne;
  RecordFileAccessorTwo: TRecordFileAccessorTwo;

implementation

{$R *.lfm}

uses
  FileUtil;

class operator TRecordOne. = (const A, B: TRecordOne): boolean;

begin
  Result := (A.ModifiedAt = B.ModifiedAt) and (A.Content = B.Content);
end;

class operator TRecordTwo. = (const A, B: TRecordTwo): boolean;

begin
  Result := (A.ModifiedAt = B.ModifiedAt) and (A.I = B.I);
end;

{ TForm1 }

procedure TForm1.ButtonAppendClick(Sender: TObject);

var
  i: integer;
  idx: integer;

begin
  i := RadioGroup1.ItemIndex;
  idx := -1;
  case i of
    0: idx := RecordFileAccessorOne.Append;
    1: idx := RecordFileAccessorTwo.Append;
    else
      assert(False, {$I %CURRENTROUTINE%} + ': undefined record type ' + IntToStr(i));
  end;
  DumpRecord(idx);
end;

procedure TForm1.ButtonMemoClearClick(Sender: TObject);
begin
  MemoEvents.Lines.Clear;
end;

procedure TForm1.ButtonDeleteClick(Sender: TObject);

var
  i: integer;
  idx: integer;

begin
  i := RadioGroup1.ItemIndex;
  idx := SpinEdit1.Value;
  case i of
    0: RecordFileAccessorOne.Delete(idx);
    1: RecordFileAccessorTwo.Delete(idx);
    else
      assert(False, {$I %CURRENTROUTINE%} + ': undefined record type ' + IntToStr(i));
  end;
  DumpRecord(idx);
end;

procedure TForm1.CheckBoxEncryptedChange(Sender: TObject);

var
  i: integer;

begin
  i := RadioGroup1.ItemIndex;
  case i of
    0: RecordFileAccessorOne.isEncrypted := TCheckBox(Sender).Checked;
    1: RecordFileAccessorTwo.isEncrypted := TCheckBox(Sender).Checked;
    else
      assert(False, {$I %CURRENTROUTINE%} + ': undefined record type ' + IntToStr(i));
  end;
end;

// --------------------------------------------------------------
// ---------------------- Auto-Refresh --------------------------
// --------------------------------------------------------------

procedure TForm1.CheckBoxAutoRefreshChange(Sender: TObject);

begin
  TimerRefresh.Interval := SpinEditRefresh.Value * 1000;
  TimerRefresh.Enabled := TCheckBox(Sender).Checked;
end;

procedure TForm1.FormActivate(Sender: TObject);
begin
  RefreshAll();
end;

procedure TForm1.TimerRefreshTimer(Sender: TObject);

begin
  RefreshAll();
end;

// ------------------------------------------------------------
// ------------------ Type Selection --------------------------
// ------------------------------------------------------------

procedure TForm1.SpinEdit1Change(Sender: TObject);

var
  idx: integer;

begin
  idx := TSpinEdit(Sender).Value;
  if CurrentAccessor.CountRec > 0 then
  begin
    if CurrentAccessor.inRange(idx, 0, CurrentAccessor.CountRec - 1) then
      DumpRecord(idx)
    else
    begin
      if idx < 0 then
        idx := CurrentAccessor.CountRec - 1
      else if idx > CurrentAccessor.CountRec - 1 then
        idx := 0;
    end;
  end
  else
    idx := -1;
  TSpinEdit(Sender).Value := idx;
end;

procedure TForm1.LabeledEditPasswordChange(Sender: TObject);
begin
  CurrentAccessor.EncryptKey := TLabeledEdit(Sender).Text;
end;

procedure TForm1.DumpRecord(idx: integer);

// Dump a record
// pass -1 to clear

begin
  // dump the cache
  Memo1.Lines.Clear;
  if CurrentAccessor.inRange(idx, 0, CurrentAccessor.CountRec - 1) then
  begin
    Memo1.Lines.Add('Cache');
    Memo1.Lines.Add(CurrentAccessor.HexDumpRec(idx));
    Memo1.Lines.Add('File');
    Memo1.Lines.Add(TBaseFileAccessor(CurrentAccessor).HexDumpRec(idx));
    StatusBar1.SimpleText := Format('%s: [%d] of %d records', [CurrentAccessor.FilePath, idx, CurrentAccessor.CountRec]);
  end
  else
    StatusBar1.SimpleText := Format(Self.ClassName + {$I %CURRENTROUTINE%} + ': %s [%d]', ['invalid index: ', idx]);
end;

function TForm1.CurrentAccessor: TBaseFileAccessor;

var
  i: integer;

begin
  i := RadioGroup1.ItemIndex;
  case i of
    0: Result := RecordFileAccessorOne;
    1: Result := RecordFileAccessorTwo;
    else
      assert(False, {$I %CURRENTROUTINE%} + ': undefined record type ' + IntToStr(i));
  end;
end;

procedure TForm1.RefreshAll;
begin
  RecordFileAccessorOne.Refresh;
  RecordFileAccessorTwo.Refresh;
end;

procedure TForm1.DisplayEvent(EventMsg: string);
begin
  MemoEvents.Lines.Insert(0, DateTimeToStr(Now()) + ': ' + EventMsg);
end;

// Event handlers
// ----------------------------------------

procedure TForm1.onAppended(Sender: TObject; idx: integer);

begin
  DisplayEvent(Format('%s.%s: %s [%d]', [Self.ClassName, {$I %CURRENTROUTINE%}, TBaseFileAccessor(Sender).FilePath, idx]));
end;

procedure TForm1.onRead(Sender: TObject; idx: integer);
begin
  DisplayEvent(Format('%s.%s: %s [%d]', [Self.ClassName, {$I %CURRENTROUTINE%}, TBaseFileAccessor(Sender).FilePath, idx]));
end;

procedure TForm1.onUpdated(Sender: TObject; idx: integer);
begin
  DisplayEvent(Format('%s.%s: %s [%d]', [Self.ClassName, {$I %CURRENTROUTINE%}, TBaseFileAccessor(Sender).FilePath, idx]));
end;

procedure TForm1.onDeleted(Sender: TObject; idx: integer);
begin
  DisplayEvent(Format('%s.%s: %s [%d]', [Self.ClassName, {$I %CURRENTROUTINE%}, TBaseFileAccessor(Sender).FilePath, idx]));
end;

procedure TForm1.onError(AClassName, ARoutineName, AFilePath: string; AnIdx: integer; AMessage: string);
begin
  DisplayEvent(Format('%s.%s: %s Error: Record [%d]: %s', [AClassName, ARoutineName, AFilePath, AnIdx, AMessage]));
end;

procedure TForm1.OnRefreshStart(Sender: TObject);
begin
  DisplayEvent(Format('%s.%s: %s', [Self.ClassName, {$I %CURRENTROUTINE%}, TBaseFileAccessor(Sender).FilePath]));
end;

procedure TForm1.OnRefreshEnd(Sender: TObject);

begin
  DisplayEvent(Format('%s.%s: %s', [Self.ClassName, {$I %CURRENTROUTINE%}, TBaseFileAccessor(Sender).FilePath]));
end;

procedure TForm1.OnRefreshed(Sender: TObject; idx: integer);
begin
  DisplayEvent(Format('%s.%s: %s [%d]', [Self.ClassName, {$I %CURRENTROUTINE%}, TBaseFileAccessor(Sender).FilePath, idx]));
end;

procedure TForm1.OnSkipped(Sender: TObject; idx: integer);
begin
  DisplayEvent(Format('%s.%s: %s [%d]', [Self.ClassName, {$I %CURRENTROUTINE%}, TBaseFileAccessor(Sender).FilePath, idx]));
end;

// ----------------------------------
// Form main routines
// ----------------------------------

procedure TForm1.FormCreate(Sender: TObject);

var
  FilePath:String;

begin
  {$IFOPT D+}
  Self.Caption := Self.Caption + ' [Debug]';
  {$ENDIF}
  DisplayEvent('GetCurrentDir: ' + Sysutils.GetCurrentDir());
  DisplayEvent('PWD: ' + GetEnvironmentVariable('PWD'));
  DisplayEvent('ParamStr(0): ' + ParamStr(0));
  DisplayEvent('ExeName: ' + Application.ExeName);
  DisplayEvent('ProgramDirectory: ' + ProgramDirectory);

  FilePath := AppendPathDelim(GetCurrentDir());
  Memo1.Lines.Clear;
  RecordFileAccessorOne := TRecordFileAccessorOne.Create(FilePath + 'RecordOne.dta');
  RecordFileAccessorOne.onAppended := @onAppended;
  RecordFileAccessorOne.onRead := @onRead;
  RecordFileAccessorOne.onUpdated := @onUpdated;
  RecordFileAccessorOne.onDeleted := @onDeleted;
  RecordFileAccessorOne.onRefreshStart := @onRefreshStart;
  RecordFileAccessorOne.onRefreshEnd := @onRefreshEnd;
  RecordFileAccessorOne.onRefreshed := @onRefreshed;
  RecordFileAccessorOne.onSkipped := @onSkipped;
  RecordFileAccessorOne.onError := @onError;
  RecordFileAccessorTwo := TRecordFileAccessorTwo.Create(FilePath + 'RecordTwo.dta');
  RecordFileAccessorTwo.onAppended := @onAppended;
  RecordFileAccessorTwo.onRead := @onRead;
  RecordFileAccessorTwo.onUpdated := @onUpdated;
  RecordFileAccessorTwo.onDeleted := @onDeleted;
  RecordFileAccessorTwo.onRefreshStart := @onRefreshStart;
  RecordFileAccessorTwo.onRefreshEnd := @onRefreshEnd;
  RecordFileAccessorTwo.onRefreshed := @onRefreshed;
  RecordFileAccessorTwo.onSkipped := @onSkipped;
  RecordFileAccessorTwo.onError := @onError;
  //
end;

procedure TForm1.FormDestroy(Sender: TObject);

begin
  FreeAndNil(RecordFileAccessorOne);
  FreeAndNil(RecordFileAccessorTwo);
end;

end.
